.nolist
#define    TASM
#include   "app.inc"
saveSScreen	= 86ECh
plotSScreen	= 9340h
OP1		= 8478h
_GrBufCpy	= 486Ah
#define		bcall(xx)	rst 28h \ .dw xx
.org 9D93h
.list
.db $BB,$6D
Start:
;disable interrupts.
;We will be using directly accessing the LCD and keys
;To make things faster, we will preload some stuff to these ports
;interrupts would mess this up
	di
;Set the LCD do Y-Decrement mode.
;For updating the LCD, we will be streaming the buffer from the lower right, reading right to left, then up.
	ld a,6
	out (16),a
;Set 15MHz if possible:
	in a,(2)
	rla
	sbc a,a
	and 3
	out (20h),a
; Now for keys, we will only be using arrows and [2nd,mode,del]
; Since these are mutually exclusive bits, we can get away with writing a value to read both key groups at once.
; Note, though, that since graph=bit 0 and down= bit 0 (for example), [graph] will be read as [down]
; If we write %10111110 to port 1, we don't need to intialise it again.
; Since no other keys are used, we can just read port 1 whenever we want key presses, no writing needed.
	ld a,%10111110	;key groups: Arrows, [graph,trace,zoom,window,y=,2nd,mode,del]
	out (1),a

	ld hl,saveSScreen
	ld de,saveSScreen+1
	ld (hl),0
	ld bc,767
	ldir
Main:
;  Check [ON] for exit
	in a,(4) \ and 8 \ ret z

	.db 21h		;ld hl,**
coords:
	.db 31,47

;  Check keys
	in a,(1)
	rlca \ call nc,Erase
	rlca \ call nc,pAdd
	rlca \ call nc,Draw
	rlca \ call nc,Adjust
	rlca \ jr c,$+3 \ dec l		;check UP
	rlca \ jr c,$+3 \ inc h		;check RIGHT
	rlca \ jr c,$+3 \ dec h		;check LEFT
	rlca \ jr c,$+3 \ inc l		;check DOWN
	ld a,l \ cp 64 \ jr nc,$+10
	ld a,h \ cp 96 \ jr nc,$+5
	ld (coords),hl	;write back the updated coordinates
	call pxlXOR
	push hl
	call Update	;updates the particles and the LCD
	pop hl
	call pxlXOR
	jp Main



;  That is it for the main loop, now come the subroutines:
;	Erase	: ([DEL])  Erase the pixel at the cursor
;	pAdd	: ([MODE]) Add a new particle to the buffer
;	Draw	: ([2nd])  Draw a pixel on the screen at the current cursor coordinates
;	pxlXOR	: Draws the cursor with XOR logic to the screen
;	Update	: Updates particles and the LCD in an interleaved routine.
;		This may have side effects such as particles not being updated until
;		after the LCD updating has already scanned past that pixel. This
;		will only cause some pixels to be rendered 1 frame later, but will still
;		look correct. As well, this gets rid of the need for an LCD delay, since
;		we can update one or two partciles per LCD write. This saves about 45000
;		clock cycles from being wasted (115000 at 15MHz), speeding up both LCD
;		updating and particle updating. This also removes the LCD as a bottleneck
;		for working at 15MHz
;	GetPixelLoc	: Returns the address where the pixel is located as well as a mask
Erase:
;need to preserve A and HL
;erase the pixel at (H,L)
;HL is always in bounds
	push af
	push hl
	ld b,h
	ld c,l
	.db 21h
	cpl
	and (hl)
	ld (PixelType),hl
	ld hl,plotSScreen
	call FastCirc

	pop hl
	pop af
	ret
Draw:
;need to preserve A and HL
;plot a pixel at (H,L)
;HL is always in bounds
	push af
	push hl
	ld b,h
	ld c,l
	.db 21h
	nop
	or (hl)
	ld (PixelType),hl
	ld hl,plotSScreen
	call FastCirc

	pop hl
	pop af
	ret
pxlXOR:
;need to preserve A and HL
;invert the pixel at (H,L)
;HL is always in bounds
	push af
	push hl
	ld b,h
	ld c,l
	.db 21h
	nop
	xor (hl)
	ld (PixelType),hl
	ld hl,plotSScreen
	call FastCirc

	pop hl
	pop af
	ret
pAdd:
;need to preserve A and HL
;plot a particle at (H,L)
;HL is always in bounds
	push af
	push hl

	ld b,h
	ld c,l
	.db 21h
	nop
	or (hl)
	ld (PixelType),hl
	ld hl,saveSScreen
	call FastCirc

	pop hl
	pop af
	ret
Adjust:
	push af
	push hl
	ld hl,(radius)
	bit 4,a
	jr nz,$+3
	dec l
	rlca
	jr c,$+3
	inc l
	ld a,l
	and 3Fh
	ld (radius),a
	pop hl
	pop af
	ret
Update:
	ld a,2Bh
	out (16),a
	ld hl,saveSScreen+767
	ld d,h
	ld e,l
	ld bc,0C00h
	in a,(16) \ rla  \ jr c,$-3
	ld a,$BF
	out (16),a
	ld ix,plotSScreen+767+12
bottomrow:
	in a,(16) \ rla  \ jr c,$-3
	ld a,(ix-12)
	or (hl)
	out (17),a
	ld (hl),c
	dec l \ dec ixl
	djnz bottomrow
	ld b,63
	ex de,hl
	in a,(16) \ rla  \ jr c,$-3
	ld a,2Bh
	out (16),a
	in a,(16) \ rla  \ jr c,$-3
loop0:
	ld a,b
	add a,7Fh
	out (16),a

	push bc
	ld b,12
	in a,(16) \ rla  \ jr c,$-3
loop:
	push bc
;move any pixels down that can
;HL points to the line of particles below
;IX points to the row below in the bitmap
;DE points to the current row
	ld a,(de)
	or (ix-12)
	out (17),a
	ld a,(ix)
	or (hl)
	cpl
	ld c,a
	ld a,(de)
	and c
	ld c,a
;C is the pixels that can move down
	or (hl)
	ld (hl),a
	ld a,(de)
	xor c
	ld (de),a
	pop bc
	dec ix
	dec de
	dec hl
	djnz loop
	
;now randomly choose a direction to shift 
; HL points to the current row
	call rand_c	;randomly sets the state of the c flag
	ld a,2Bh
	out (16),a
	call c,rowright	;must preserve all registers
	call nc,rowleft	;must preserve all registers
;	ccf
;	call nc,rowright;must preserve all registers
;	call c,rowleft	;must preserve all registers
;save the current state of the row to AppBackUpScreen
;after all of this is done, we need to see if those pixels can be moved up
	pop bc
	djnz loop0
	ret


rand_c:
	push hl
;  From CaDan, by Iambian and Geekboy
LFSR2Byte:
	.db 21h
lfsrseed2:
	.dw 2
  ld a,L
  rlca
  xor h
  rlca  ;
  rlca  ;
  xor h ;
  rlca
  xor h
  rlca
  adc hl,hl
  ld (lfsrseed2),hl
	pop hl
	ret
rowright:

	push af
	push ix
	push hl
	push de
	ld h,d
	ld l,e
	ld de,OP1-1
	ld b,12
	or a
	inc hl \ inc e \ ld a,(hl) \ rra \ ld (de),a \ djnz $-5
	ex de,hl
	
;IX points to the main screen
;HL points to the shifted data
;DE points to non-shifted data
	ld b,12
	or a
	ex af,af'
rrloop1:
	ld a,(de) \ or (ix-12) \ cpl \ and (hl)
	ld c,a
;A has the pixels that can move
	ex de,hl \ xor (hl) \ ld (hl),a
	ex af,af' \ ld a,c \ rla \ ld c,a \ ex af,af'
	ld a,c
	xor (hl) \ ld (hl),a
	ex de,hl
	dec l \ dec de \ dec ix \ djnz rrloop1

	pop de
	pop hl
	pop ix
	pop af
	ret
rowleft:
	push af
	push ix
	push hl
	push de
	ld de,OP1+11
	ld b,12
	or a
	ld a,(hl) \ rla \ ld (de),a \ dec e \ dec hl \ djnz $-5
	ex de,hl
	
;IX points to the main screen
;HL points to the shifted data
;DE points to non-shifted data
	ld b,12
	or a
	ex af,af'
rlloop1:
	inc l \ inc de \ inc ix
	ld a,(de) \ or (ix-12) \ cpl \ and (hl)
	ld c,a
	ex de,hl \ xor (hl) \ ld (hl),a
	ex af,af' \ ld a,c \ rra \ ld c,a \ ex af,af'
	ld a,c \ xor (hl) \ ld (hl),a
	ex de,hl
	djnz rlloop1

	pop de
	pop hl
	pop ix
	pop af
	ret


FastCirc:
;Inputs:
;     BC = (x,y)
	ld (buff),hl

	.db 26h
radius:
	.db 2
     ld a,h
     or a
     ret z
     ret m
     ld e,0
     ld d,a
     ld l,e
     add hl,hl
     jp cLoop+4
cLoop:
     ex af,af'
     call Plot4Pix
     call Plot4Pix
     inc d
     sub l
     jr nc,YIsGood
DecY:
     dec e
     inc l
     add a,h
YIsGood:
     ex af,af'
     ld a,e
     sub d
     jp p,cLoop
     ret

Plot4Pix:
     inc l
     push af
     push hl
     push bc
     ld a,e
     ld e,d
     ld d,a
     push de
     add a,b
     cp 96
     call c,Plot2Pix       ;DC****
     ld a,b
     sub d
     call p,Plot2Pix
     pop de
     pop bc
     pop hl
     pop af
     ret
plot2Pix:
     push bc
     ld b,a
     push bc
     ld a,c
     add a,e
     ld c,a
;c+e
     call p,PlotPixel
     pop bc
     ld a,c
     sub e
     ld c,a
;c-e
     call p,PlotPixel
     pop bc
     ret

PlotPixel:
;Input:
;     b is X
;     c,a is y
;	DE is the buf
;Output:
;     HL points to byte
;     A is the mask
;     nc if not computed, c if computed

       cp 64 \ ret nc
       ld a,b
       cp 96 \ ret nc
       ld l,c
       ld h,0
       ld b,h
       sla l
       add hl,bc
       add hl,hl
       add hl,hl
       ld b,a
       rrca \ rrca \ rrca
       and 0Fh
       ld c,a
       ld a,b
	ld b,0
       add hl,bc
	.db 01h
buff:
	.dw saveSScreen
	add hl,bc
       and 7
       ld b,a
       inc b
       ld a,1
         rrca
         djnz $-1
PixelType:
       nop
       or (hl)
       ld (hl),a
       ret